from __future__ import annotations

import asyncio
import datetime
from typing import Any, TypeGuard, TypeVar, cast

from polyfactory.factories.pydantic_factory import ModelFactory
from polyfactory.persistence import AsyncPersistenceProtocol

from gcp_pilot.firestore.document import Document

T = TypeVar("T", bound=Document)


def _convert_dates_to_datetimes(data: Any) -> Any:
    if isinstance(data, dict):
        return {key: _convert_dates_to_datetimes(value) for key, value in data.items()}
    if isinstance(data, list):
        return [_convert_dates_to_datetimes(item) for item in data]
    if isinstance(data, datetime.date) and not isinstance(data, datetime.datetime):
        return datetime.datetime.combine(data, datetime.time.min)
    return data


class FirestorePersistenceHandler(AsyncPersistenceProtocol[T]):
    async def save(self, data: T) -> T:
        doc_data = data.model_dump(by_alias=True, exclude={"id"})
        converted_data = _convert_dates_to_datetimes(doc_data)
        created_doc = await data.__class__.documents.create(data=converted_data)
        return cast(T, created_doc)

    async def save_many(self, data: list[T]) -> list[T]:
        # TODO: This can be optimized to use a single batch operation
        return await asyncio.gather(*[self.save(item) for item in data])


class FirestoreFactory(ModelFactory[T]):
    __is_async__ = True
    __async_persistence__ = FirestorePersistenceHandler
    __is_base_factory__ = True

    @classmethod
    def is_supported_type(cls, value: Any) -> TypeGuard[type[T]]:
        return issubclass(value, Document)



import pytest
from google.cloud.firestore import AsyncClient

from gcp_pilot.firestore import DoesNotExist
from tests.gcp_pilot.firestore_test.conftest import Product


@pytest.mark.asyncio
class TestBasicFirestoreORM:
    async def test_client_is_async(self):
        assert isinstance(Product.documents.client, AsyncClient)

    async def test_document_create(self, product_data):
        product = Product(**product_data)
        assert product.pk is None

        product = await product.save()
        assert product.pk is not None
        assert product.name == product_data["name"]

    async def test_document_get(self, saved_product):
        product = await Product.documents.get(pk=saved_product.pk)
        assert product.pk == saved_product.pk
        assert product.name == saved_product.name

    async def test_document_save_update(self, saved_product):
        new_name = "Updated Product Name"
        saved_product.name = new_name
        await saved_product.save()

        updated_product = await Product.documents.filter(name__eq=new_name).get()
        assert updated_product.name == new_name

    async def test_document_delete(self, saved_product):
        pk = saved_product.pk
        await saved_product.delete()

        with pytest.raises(DoesNotExist):
            await Product.documents.get(pk=pk)



from contextlib import ExitStack
from functools import wraps
from unittest.mock import Mock, patch

from google.oauth2.service_account import Credentials


class patch_auth:
    def __init__(self, project_id: str = "potato-dev", location: str = "moon-dark1", email: str = "chuck@norris.com"):
        # Realistic: actual class to be accepted by clients during validation
        # But fake: with as few attributes as possible, any API call using the credential should fail
        credentials = Credentials(
            service_account_email=email,
            signer=None,
            token_uri="",
            project_id=project_id,
        )
        managers = [
            patch("google.auth.default", return_value=(credentials, project_id)),
            patch("google.oauth2.service_account.Credentials.before_request", return_value=None),
            patch("google.oauth2.service_account.Credentials.refresh", return_value=None),
            patch("gcp_pilot.base.GoogleCloudPilotAPI._set_location", return_value=location),
            patch("gcp_pilot.base.AppEngineBasedService._set_location", return_value=location),
        ]
        self.stack = ExitStack()
        for mgr in managers:
            self.stack.enter_context(mgr)

    def __enter__(self):
        return self.stack.__enter__()

    def start(self):
        return self.__enter__()

    def __exit__(self, typ, val, traceback):
        return self.stack.__exit__(typ, val, traceback)

    def stop(self):
        self.__exit__(None, None, None)

    def __call__(self, func):
        @wraps(func)
        def wrapper(*args, **kw):
            with self:
                return func(*args, **kw)

        return wrapper


def patch_firebase_token(**kwargs):
    return patch("google.oauth2.id_token.verify_firebase_token", **kwargs)


class DiscoveryMixinTest:
    def setUp(self):
        self.mocked_discovery_client = Mock()
        _patch_build_client = patch(
            "gcp_pilot.base.GoogleCloudPilotAPI._build_client", return_value=self.mocked_discovery_client
        )
        _patch_build_client.start()
        self.addCleanup(_patch_build_client.stop)



from __future__ import annotations

import abc
from typing import TYPE_CHECKING, Any, ClassVar

from pydantic import BaseModel, ConfigDict, Field, PrivateAttr

from gcp_pilot.firestore.subcollection import Subcollection

if TYPE_CHECKING:
    from gcp_pilot.firestore.manager import Manager


DEFAULT_PK_FIELD_NAME = "id"


class Options:
    def __init__(self, meta, klass: type[Document]):
        self.meta = meta
        self.model_class = klass
        self.collection_name: str = getattr(meta, "collection_name", klass.__name__.lower())
        self.pk_field_name: str = self._get_pk_field_name()

    def _get_pk_field_name(self) -> str:
        pk_field = [
            name for name, field in self.model_class.model_fields.items() if field.alias == DEFAULT_PK_FIELD_NAME
        ]
        if not pk_field:
            raise TypeError("No primary key field found.")
        return pk_field[0]


class Document(BaseModel, abc.ABC):
    id: str | None = Field(default=None, alias="id")
    _manager: Manager | None = PrivateAttr(default=None)

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        use_enum_values=True,
        ignored_types=(Subcollection,),
    )

    if TYPE_CHECKING:
        documents: ClassVar[Manager]
        _meta: ClassVar[Options]

    def __init_subclass__(cls, **kwargs):
        from gcp_pilot.firestore.manager import Manager  # noqa: PLC0415

        super().__init_subclass__(**kwargs)
        manager = Manager(doc_klass=cls)
        cls.documents = manager
        meta = getattr(cls, "Meta", None)
        cls._meta = Options(meta=meta, klass=cls)

    def __init__(self, **data: Any):
        super().__init__(**data)
        self._manager = self.__class__.documents

    @property
    def pk(self) -> str | None:
        return self.id

    @property
    def manager(self) -> Manager:
        return self._manager or self.__class__.documents

    async def save(self) -> Document:
        data = self.model_dump(mode="json", by_alias=True, exclude={self._meta.pk_field_name})

        saved_doc = await self.manager.create(data=data, pk=self.pk)
        self.id = saved_doc.pk
        return self

    async def update(self, **kwargs: Any) -> None:
        if not self.pk:
            raise ValueError("Cannot update a document without a primary key.")
        await self.manager.update(pk=self.pk, data=kwargs)

    async def refresh(self) -> None:
        if not self.pk:
            raise ValueError("Cannot refresh a document without a primary key.")
        refreshed_doc = await self.manager.get(pk=self.pk)
        object_attributes = refreshed_doc.model_dump()
        for key, value in object_attributes.items():
            setattr(self, key, value)

    async def delete(self) -> None:
        if not self.pk:
            raise ValueError("Cannot delete a document without a primary key.")
        await self.manager.delete(pk=self.pk)

    def __repr__(self):
        return f"<{self.__class__.__name__} pk={self.pk}>"



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from google.cloud import firestore
from google.cloud.firestore_v1.async_client import AsyncClient
from google.cloud.firestore_v1.async_collection import AsyncCollectionReference

from gcp_pilot.firestore.atomic import _active_batch
from gcp_pilot.firestore.exceptions import DoesNotExist
from gcp_pilot.firestore.query import Query

if TYPE_CHECKING:
    from gcp_pilot.firestore.document import Document


class Manager:
    _client: AsyncClient | None = None

    def __init__(self, doc_klass: type[Document], parent: Document | None = None):
        self.doc_klass = doc_klass
        self.parent = parent

    def _to_document(self, data: dict[str, Any]) -> Document:
        document = self.doc_klass.model_validate(data)
        document._manager = self
        return document

    @property
    def client(self) -> AsyncClient:
        if not self.__class__._client:
            self.__class__._client = firestore.AsyncClient()
        return self.__class__._client  # type: ignore

    @property
    def collection_name(self) -> str:
        return self.doc_klass._meta.collection_name

    @property
    def collection(self) -> AsyncCollectionReference:
        if not self.parent:
            return self.client.collection(self.collection_name)

        if not self.parent.pk:
            raise ValueError("Cannot access subcollection on a document without a primary key.")

        return self.parent.documents.collection.document(self.parent.pk).collection(self.collection_name)

    def _get_query(self) -> Query:
        return Query(manager=self)

    async def get(self, pk: str | None = None, **kwargs) -> Document:
        if pk and kwargs:
            raise ValueError("Cannot use pk and kwargs together.")

        if pk:
            doc_ref = self.collection.document(pk)
            doc_snapshot = await doc_ref.get()
            if not doc_snapshot.exists:
                raise DoesNotExist(self.doc_klass, {"id": pk})

            data = doc_snapshot.to_dict() or {}
            data["id"] = doc_snapshot.id
            return self._to_document(data)

        return await self.filter(**kwargs).get()

    async def create(self, data: dict[str, Any], pk: str | None = None) -> Document:
        doc_ref = self.collection.document(pk)
        document = self._to_document({"id": doc_ref.id, **data})

        batch = _active_batch.get()
        if batch is not None:
            batch.set(doc_ref, data)
        else:
            await doc_ref.set(data)

        return document

    async def update(self, pk: str, data: dict[str, Any]) -> None:
        doc_ref = self.collection.document(pk)
        batch = _active_batch.get()
        if batch is not None:
            batch.update(doc_ref, data)
        else:
            await doc_ref.update(data)

    async def delete(self, pk: str) -> None:
        doc_ref = self.collection.document(pk)
        batch = _active_batch.get()
        if batch is not None:
            batch.delete(doc_ref)
        else:
            await doc_ref.delete()

    def __getattr__(self, name: str) -> Any:
        return getattr(self._get_query(), name)
